#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "trimesh.h"

void create_initial_circle_mesh(TriMesh *mesh, float spacing, float radius, float height);
void snap_boundary_nodes_to_circle(TriMesh *mesh, float radius);
void smooth_interior_nodes(TriMesh *mesh);

int main(int argc, char **argv)
{
   TriMesh mesh;
   float spacing=0.1, radius=1, height=1;

   if(argc<2 || argc>5){
      fprintf(stderr, "usage: circle_mesh <output_filename> [spacing] [radius] [height]\n");
      return 1;
   }

   if(argc>2) sscanf(argv[2], "%f", &spacing);
   if(argc>3) sscanf(argv[3], "%f", &radius);
   if(argc>4) sscanf(argv[4], "%f", &height);

   create_initial_circle_mesh(&mesh, spacing, radius, height);
   compute_adjacencies(&mesh);
   snap_boundary_nodes_to_circle(&mesh, radius);
   smooth_interior_nodes(&mesh);
   smooth_interior_nodes(&mesh);
   write_tri_mesh(argv[1], &mesh);

   return 0;
}

static inline int circle_check(float x, float z, float radius)
{ return x*x+z*z<=radius*radius; }

void create_initial_circle_mesh(TriMesh *mesh, float spacing, float radius, float height)
{
   int nx, nz, i, j, *indices, t;
   float dx, dz;

   dx=spacing;
   dz=spacing*sqrt(3)/2;
   nx=2+(int)ceil(2*radius/dx);
   nz=1+(int)ceil(2*radius/dz);
   indices=(int*)malloc(nx*nz*sizeof(int));

   #define index(a,b) (indices[(b)+nz*(a)])
   #define gridx(a,b) (((a)-0.5*(1-(b)%2))*dx-radius)
   #define gridz(a,b) ((b)*dz-radius)

   // first mark all grid indices as unused
   for(i=0; i<nx; ++i){
      for(j=0; j<nz; ++j){
         index(i,j)=-1;
      }
   }

   // then loop through grid triangles to see which overlap the circle enough
   // (while we're at it, count the number of final mesh vertices and triangles)
   mesh->nv=0;
   mesh->nt=0;
   for(i=0; i<nx-1; ++i){
      for(j=0; j<nz-1; ++j){
         if(j%2){
            // check on triangle (i,j)-(i+1,j+1)-(i,j+1)
            if(circle_check(gridx(i,j), gridz(i,j), radius)
            && circle_check(gridx(i+1,j+1), gridz(i+1,j+1), radius)
            && circle_check(gridx(i,j+1), gridz(i,j+1), radius)){
               if(index(i,j)<0)     {index(i,j)=mesh->nv;     ++mesh->nv;}
               if(index(i+1,j+1)<0) {index(i+1,j+1)=mesh->nv; ++mesh->nv;}
               if(index(i,j+1)<0)   {index(i,j+1)=mesh->nv;   ++mesh->nv;}
               ++mesh->nt;
            }
            // check on triangle (i,j)-(i+1,j)-(i+1,j+1)
            if(circle_check(gridx(i,j), gridz(i,j), radius)
            && circle_check(gridx(i+1,j), gridz(i+1,j), radius)
            && circle_check(gridx(i+1,j+1), gridz(i+1,j+1), radius)){
               if(index(i,j)<0)     {index(i,j)=mesh->nv;     ++mesh->nv;}
               if(index(i+1,j)<0)   {index(i+1,j)=mesh->nv;   ++mesh->nv;}
               if(index(i+1,j+1)<0) {index(i+1,j+1)=mesh->nv; ++mesh->nv;}
               ++mesh->nt;
            }
         }else{
            // check on triangle (i,j)-(i+1,j)-(i,j+1)
            if(circle_check(gridx(i,j), gridz(i,j), radius)
            && circle_check(gridx(i+1,j), gridz(i+1,j), radius)
            && circle_check(gridx(i,j+1), gridz(i,j+1), radius)){
               if(index(i,j)<0)   {index(i,j)=mesh->nv;   ++mesh->nv;}
               if(index(i+1,j)<0) {index(i+1,j)=mesh->nv; ++mesh->nv;}
               if(index(i,j+1)<0) {index(i,j+1)=mesh->nv; ++mesh->nv;}
               ++mesh->nt;
            }
            // check on triangle (i+1,j)-(i+1,j+1)-(i,j+1)
            if(circle_check(gridx(i+1,j), gridz(i+1,j), radius)
            && circle_check(gridx(i+1,j+1), gridz(i+1,j+1), radius)
            && circle_check(gridx(i,j+1), gridz(i,j+1), radius)){
               if(index(i+1,j)<0)   {index(i+1,j)=mesh->nv;   ++mesh->nv;}
               if(index(i+1,j+1)<0) {index(i+1,j+1)=mesh->nv; ++mesh->nv;}
               if(index(i,j+1)<0)   {index(i,j+1)=mesh->nv;   ++mesh->nv;}
               ++mesh->nt;
            }
         }
      }
   }

   mesh->x=(float*)malloc(3*mesh->nv*sizeof(float));
   mesh->tri=(int*)malloc(3*mesh->nt*sizeof(int));

   // now create the vertices
   for(i=0; i<nx; ++i){
      for(j=0; j<nz; ++j){
         if(index(i,j)>=0){
            mesh->x[3*index(i,j)]=gridx(i,j);
            mesh->x[3*index(i,j)+1]=height;
            mesh->x[3*index(i,j)+2]=gridz(i,j);
         }
      }
   }

   // and create the triangles
   t=0;
   for(i=0; i<nx-1; ++i){
      for(j=0; j<nz-1; ++j){
         if(j%2){
            // check on triangle (i,j)-(i+1,j+1)-(i,j+1)
            if(index(i,j)>=0 && index(i+1,j+1)>=0 && index(i,j+1)>=0){
               mesh->tri[3*t]=index(i,j);
               mesh->tri[3*t+1]=index(i+1,j+1);
               mesh->tri[3*t+2]=index(i,j+1);
               ++t;
            }
            // check on triangle (i,j)-(i+1,j)-(i+1,j+1)
            if(index(i,j)>=0 && index(i+1,j)>=0 && index(i+1,j+1)>=0){
               mesh->tri[3*t]=index(i,j);
               mesh->tri[3*t+1]=index(i+1,j);
               mesh->tri[3*t+2]=index(i+1,j+1);
               ++t;
            }
         }else{
            // check on triangle (i,j)-(i+1,j)-(i,j+1)
            if(index(i,j)>=0 && index(i+1,j)>=0 && index(i,j+1)>=0){
               mesh->tri[3*t]=index(i,j);
               mesh->tri[3*t+1]=index(i+1,j);
               mesh->tri[3*t+2]=index(i,j+1);
               ++t;
            }
            // check on triangle (i+1,j)-(i+1,j+1)-(i,j+1)
            if(index(i+1,j)>=0 && index(i+1,j+1)>=0 && index(i,j+1)>=0){
               mesh->tri[3*t]=index(i+1,j);
               mesh->tri[3*t+1]=index(i+1,j+1);
               mesh->tri[3*t+2]=index(i,j+1);
               ++t;
            }
         }
      }
   }

   free(indices);
   #undef index
   #undef gridx
   #undef gridy
}

void snap_boundary_nodes_to_circle(TriMesh *mesh, float radius)
{
   int i;
   float mag;
   for(i=0; i<mesh->nv; ++i){
      if(mesh->bdry[i]){
         mag=sqrt(mesh->x[3*i]*mesh->x[3*i]+mesh->x[3*i+2]*mesh->x[3*i+2]);
         mesh->x[3*i]*=radius/mag;
         mesh->x[3*i+2]*=radius/mag;
      }
   }
}

void smooth_interior_nodes(TriMesh *mesh)
{
   int i, j;
   for(i=0; i<mesh->nv; ++i){
      if(!mesh->bdry[i]){
         mesh->x[3*i]=0;
         mesh->x[3*i+2]=0;
         for(j=0; j<mesh->deg[i]; ++j){
            mesh->x[3*i]+=mesh->x[3*mesh->adj[i][j]];
            mesh->x[3*i+2]+=mesh->x[3*mesh->adj[i][j]+2];
         }
         mesh->x[3*i]/=mesh->deg[i];
         mesh->x[3*i+2]/=mesh->deg[i];
      }
   }
   for(i=mesh->nv-2; i>=0; --i){
      if(!mesh->bdry[i]){
         mesh->x[3*i]=0;
         mesh->x[3*i+2]=0;
         for(j=0; j<mesh->deg[i]; ++j){
            mesh->x[3*i]+=mesh->x[3*mesh->adj[i][j]];
            mesh->x[3*i+2]+=mesh->x[3*mesh->adj[i][j]+2];
         }
         mesh->x[3*i]/=mesh->deg[i];
         mesh->x[3*i+2]/=mesh->deg[i];
      }
   }
}

