#include <stdio.h>
#include <stdlib.h>
#include "trimesh.h"

void read_tri_mesh(const char *filename, TriMesh *mesh)
{
   FILE *fp;
   int i;

   fp=fopen(filename, "rt");
   if(!fp) return;
   fscanf(fp, "%d", &mesh->nv);
   mesh->x=(float*)malloc(3*mesh->nv*sizeof(float));
   for(i=0; i<3*mesh->nv; ++i)
      fscanf(fp, "%g", mesh->x+i);
   fscanf(fp, "%d", &mesh->nt);
   mesh->tri=(int*)malloc(3*mesh->nt*sizeof(int));
   for(i=0; i<3*mesh->nt; ++i)
      fscanf(fp, "%d", mesh->tri+i);
   fclose(fp);

   mesh->ne=0;
   mesh->nb=0;
   mesh->edge=0;
   mesh->bend=0;
}

void write_tri_mesh(const char *filename, TriMesh *mesh)
{
   FILE *fp;
   int i;

   fp=fopen(filename, "wt");
   if(!fp) return;
   fprintf(fp, "%d\n", mesh->nv);
   for(i=0; i<mesh->nv; ++i)
      fprintf(fp, "%.7g %.7g %.7g\n", mesh->x[3*i], mesh->x[3*i+1], mesh->x[3*i+2]);
   fprintf(fp, "%d\n", mesh->nt);
   for(i=0; i<mesh->nt; ++i)
      fprintf(fp, "%d %d %d\n", mesh->tri[3*i], mesh->tri[3*i+1], mesh->tri[3*i+2]);
   fclose(fp);
}

static void add_occurrences(const int *tri, const int *trilist, int listlen, int vertex, int *ne, int *nb)
{
   int p, t, m=0;

   for(p=0; p<listlen; ++p){
      t=trilist[p];
      if(tri[3*t]==vertex || tri[3*t+1]==vertex || tri[3*t+2]==vertex)
         ++m;
   }
   if(m==1){
      *ne+=2;
   }else if(m==2){
      *ne+=1;
      *nb+=1;
   }else{
      fprintf(stderr, "triangle mesh is not manifold (edge on vertex %d shared by %d triangles)\n", vertex, m);
      exit(2);
   }
}

static void append_unique_edge(int *list, int *count, int common_endpoint, int new_endpoint)
{
   int i;
   for(i=0; i<*count; ++i)
      if(list[2*i+1]==new_endpoint) return;
   list[2*(*count)]=common_endpoint;
   list[2*(*count)+1]=new_endpoint;
   ++*count;
}

// only return true and set k if the adjacent triangle has index larger than t
static int find_adjacent_triangle(int *tri, int t, int *trilist, int len, int i, int j, int *k)
{
   int p, s;
   for(p=0; p<len; ++p){
      s=trilist[p];
      if(s>t){
         if(tri[3*s]==i && tri[3*s+1]==j){
            *k=tri[3*s+2];
            return 1;
         }else if(tri[3*s]==i && tri[3*s+2]==j){
            *k=tri[3*s+1];
            return 1;
         }else if(tri[3*s+1]==i && tri[3*s+2]==j){
            *k=tri[3*s];
            return 1;
         }else if(tri[3*s]==j && tri[3*s+1]==i){
            *k=tri[3*s+2];
            return 1;
         }else if(tri[3*s]==j && tri[3*s+2]==i){
            *k=tri[3*s+1];
            return 1;
         }else if(tri[3*s+1]==j && tri[3*s+2]==i){
            *k=tri[3*s];
            return 1;
         }
      }
   }
   return 0;
}

void compute_auxiliary_structures(TriMesh *mesh)
{
   int i, j, k, t, estart, elen, bcount;
   int *tricount, *tc;
   int **inctri; // incident triangles

   // first get a list of incident triangles for each vertex
   tricount=(int*)calloc(mesh->nv, sizeof(int));
   for(t=0; t<mesh->nt; ++t){
      ++tricount[mesh->tri[3*t]];
      ++tricount[mesh->tri[3*t+1]];
      ++tricount[mesh->tri[3*t+2]];
   }
   inctri=(int**)malloc(mesh->nv*sizeof(int*));
   for(i=0; i<mesh->nv; ++i)
      inctri[i]=(int*)malloc(tricount[i]*sizeof(int));
   tc=(int*)calloc(mesh->nv, sizeof(int));
   for(t=0; t<mesh->nt; ++t){
      i=mesh->tri[3*t]; j=mesh->tri[3*t+1]; k=mesh->tri[3*t+2]; 
      inctri[i][tc[i]]=t; ++tc[i];
      inctri[j][tc[j]]=t; ++tc[j];
      inctri[k][tc[k]]=t; ++tc[k];
   }
   free(tc);

   // next figure out how many edges and how many bending elements (interior edges) there are
   mesh->ne=0;
   mesh->nb=0;
   for(i=0; i<mesh->nv; ++i){
      for(j=0; j<tricount[i]; ++j){
         t=inctri[i][j];
         if(mesh->tri[3*t]>i)
            add_occurrences(mesh->tri, inctri[i], tricount[i], mesh->tri[3*t], &mesh->ne, &mesh->nb);
         if(mesh->tri[3*t+1]>i)
            add_occurrences(mesh->tri, inctri[i], tricount[i], mesh->tri[3*t+1], &mesh->ne, &mesh->nb);
         if(mesh->tri[3*t+2]>i)
            add_occurrences(mesh->tri, inctri[i], tricount[i], mesh->tri[3*t+2], &mesh->ne, &mesh->nb);
      }
   }
   // we overcounted by exactly a factor of two; correct it now
   mesh->ne/=2;
   mesh->nb/=2;

   // get storage for the list of edges and bending elements
   mesh->edge=(int*)malloc(2*mesh->ne*sizeof(int));
   mesh->bend=(int*)malloc(4*mesh->nb*sizeof(int));
  
   // fill the edge list in 
   estart=0;
   for(i=0; i<mesh->nv; ++i){
      elen=0;
      for(j=0; j<tricount[i]; ++j){
         t=inctri[i][j];
         if(mesh->tri[3*t]>i)
            append_unique_edge(mesh->edge+2*estart, &elen, i, mesh->tri[3*t]);
         if(mesh->tri[3*t+1]>i)
            append_unique_edge(mesh->edge+2*estart, &elen, i, mesh->tri[3*t+1]);
         if(mesh->tri[3*t+2]>i)
            append_unique_edge(mesh->edge+2*estart, &elen, i, mesh->tri[3*t+2]);
      }
      estart+=elen;
   }

   // fill the bend list in
   bcount=0;
   for(t=0; t<mesh->nt; ++t){
      int h;
      i=mesh->tri[3*t]; j=mesh->tri[3*t+1]; k=mesh->tri[3*t+2];
      if(find_adjacent_triangle(mesh->tri, t, inctri[i], tricount[i], i, j, &h)){
         mesh->bend[4*bcount]=k;
         mesh->bend[4*bcount+1]=i;
         mesh->bend[4*bcount+2]=j;
         mesh->bend[4*bcount+3]=h;
         ++bcount;
      }
      if(find_adjacent_triangle(mesh->tri, t, inctri[j], tricount[j], j, k, &h)){
         mesh->bend[4*bcount]=i;
         mesh->bend[4*bcount+1]=j;
         mesh->bend[4*bcount+2]=k;
         mesh->bend[4*bcount+3]=h;
         ++bcount;
      }
      if(find_adjacent_triangle(mesh->tri, t, inctri[k], tricount[k], k, i, &h)){
         mesh->bend[4*bcount]=j;
         mesh->bend[4*bcount+1]=k;
         mesh->bend[4*bcount+2]=i;
         mesh->bend[4*bcount+3]=h;
         ++bcount;
      }
   }

   free(tricount);
   free(inctri);
}

