#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "trimesh.h"

// globals for rendering
int fullscreen_mode=0;
int width=800, height=600;
int saved_width, saved_height;
int oldmousex=0, oldmousey=0;
enum {DONOTHING, ROTATE, ZOOM, MOVETARGET} mousemode=DONOTHING;
float camdist=6;
const float NEARCLIP_FACTOR=0.1, FARCLIP_FACTOR=10;
float nearclip, farclip;
float target[3]={0, 1, 0};
float view_target_timer=0;
int view_ground=1, view_buddha=1;
float old_time=0;
const float fovy=45;
float heading=M_PI/3, pitch=0;
float campos[3];
int file_count=0;
int wireframe_mode=0;

// globals for data
int frame=0;
const char *directory=0;
TriMesh cloth;
TriMesh object;

// prototypes
void read_initial();
void read_frame(int new_frame);
void handle_display();
void handle_reshape(int w, int h);
void handle_keypress(unsigned char key, int x, int y);
void handle_click(int button, int state, int x, int y);
void handle_drag(int x, int y);
void handle_idle();
void toggle_fullscreen();
void compute_camera_position();
void display_ground();
void display_mesh(const TriMesh *mesh);
void display_target();
void save_screen(const char *filename);
void init_lighting();
void init_material(GLenum face, float r, float g, float b, float specularity);

// definitions
int main(int argc, char *argv[])
{
   // initialize GLUT stuff 
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowSize(width,height);
   saved_width=width;
   saved_height=height;
   glutCreateWindow("cloth viewer (hw4)");
   glutDisplayFunc(handle_display);
   glutReshapeFunc(handle_reshape);
   glutKeyboardFunc(handle_keypress);
   glutMouseFunc(handle_click);
   glutMotionFunc(handle_drag);
   glutIdleFunc(handle_idle);

   // initialize OpenGL stuff
   glEnable(GL_DEPTH_TEST);
   glClearColor(0, 0, 0, 0);
   glClearDepth(1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   nearclip=NEARCLIP_FACTOR*camdist;
   farclip=FARCLIP_FACTOR*camdist;
   init_lighting();
   glEnable(GL_LIGHTING);
   glDisable(GL_CULL_FACE);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

   if(argc<2){
      fprintf(stderr, "Expecting <cloth_output_directory> as argument\n");
      return 1;
   }
   directory=argv[1];

   // initialize first frame
   read_initial();

   // let's go
   old_time=clock()/(float)CLOCKS_PER_SEC;
   glutPostRedisplay();
   glutMainLoop();
   return 0;
}

void init_lighting()
{
   {
      GLfloat ambient[4] = {.3, .3, .3, 1};
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
   }{
      GLfloat color[4] = {.7, .7, .7, 1};
      glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
      glLightfv(GL_LIGHT0, GL_SPECULAR, color);
      glEnable(GL_LIGHT0);
   }{
      GLfloat color[4] = {.4, .4, .4, 1};
      glLightfv(GL_LIGHT1, GL_DIFFUSE, color);
      glLightfv(GL_LIGHT1, GL_SPECULAR, color);
      glEnable(GL_LIGHT1);
   }{
      GLfloat color[4] = {.2, .2, .2, 1};
      glLightfv(GL_LIGHT2, GL_DIFFUSE, color);
      glLightfv(GL_LIGHT2, GL_SPECULAR, color);
      glEnable(GL_LIGHT2);
   }
}

void init_material(GLenum face, float r, float g, float b, float specularity)
{
   GLfloat f_ambient[4]={0}, f_diffuse[4]={0}, f_specular[4]={0};

   f_ambient[0]=0.1*r;
   f_ambient[1]=0.1*g;
   f_ambient[2]=0.1+0.2*b;
   f_diffuse[0]=(0.6+0.3*(1-specularity))*r;
   f_diffuse[1]=(0.6+0.3*(1-specularity))*g;
   f_diffuse[2]=(0.4+0.3*(1-specularity))*b;
   f_specular[0]=0.3*specularity;
   f_specular[1]=0.3*specularity;
   f_specular[2]=0.2*specularity;

   glMaterialfv(face, GL_AMBIENT, f_ambient);
   glMaterialfv(face, GL_DIFFUSE, f_diffuse);
   if(specularity){
      glMaterialfv(face, GL_SPECULAR, f_specular);
      glMaterialf(face, GL_SHININESS, 64);
   }
}

void read_initial()
{
   char *filename;

   filename=malloc(strlen(directory)+30);
   sprintf(filename, "%s/cloth0000.mesh", directory);
   read_tri_mesh(filename, &cloth);
   compute_vertex_normals(&cloth);
   sprintf(filename, "%s/object.mesh", directory);
   read_tri_mesh(filename, &object);
   compute_vertex_normals(&object);
   free(filename); 
}

void read_frame(int new_frame)
{
   char *filename;
   FILE *fp;
   int i, nv;

   if(new_frame<0) return;
   filename=malloc(strlen(directory)+30);
   sprintf(filename, "%s/cloth%04d.mesh", directory, new_frame);
   fp=fopen(filename, "rt");
   if(fp){
      fscanf(fp, "%d", &nv);
      if(nv==cloth.nv){
         frame=new_frame;
         for(i=0; i<3*nv; ++i)
            fscanf(fp, "%f", cloth.x+i);
         compute_vertex_normals(&cloth);
      }
      fclose(fp);
   }
   free(filename);
}

void handle_display()
{
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(campos[0], campos[1], campos[2], target[0], target[1], target[2], 0, 1, 0);

   if(view_ground){
      init_material(GL_FRONT_AND_BACK, 0.7, 0.7, 0.7, 0);
      display_ground();
   }
   init_material(GL_FRONT_AND_BACK, 0.7, 0.7, 0.2, 1);
   display_mesh(&object);
   if(wireframe_mode) glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   init_material(GL_FRONT, 0.1, 0.8, 0.9, 0.5);
   init_material(GL_BACK, 0.9, 0.1, 0.8, 0.5);
   display_mesh(&cloth);
   if(wireframe_mode) glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   if(view_target_timer>0) display_target();

   glutSwapBuffers();
}

void handle_reshape(int w, int h)
{
   width=w;
   height=h;
   glViewport(0, 0, (GLsizei)width, (GLsizei)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy, width/(float)height, nearclip, farclip);
   glutPostRedisplay();
}

void handle_keypress(unsigned char key, int x, int y)
{
   switch(key){
   case '.':
      read_frame(frame+1);
      break;
   case ',':
      read_frame(frame-1);
      break;
   case 'f':
      toggle_fullscreen();
      break;
   case 'g':
      view_ground=!view_ground;
      break;
   case 'q':
      exit(0);
   case 's':
      {
         char filename[30];
         sprintf(filename, "hw4_%04d.ppm", file_count);
         save_screen(filename);
         ++file_count;
      }
      break;
   case 'w':
      wireframe_mode=!wireframe_mode;
      break;
   default:
      return;
   }
   glutPostRedisplay();
}

void handle_click(int button, int state, int x, int y)
{
   int mods=glutGetModifiers();
   if(state==GLUT_UP) mousemode = DONOTHING;
   else if(mods&GLUT_ACTIVE_SHIFT){
      view_target_timer=1;
      mousemode=MOVETARGET;
   }else{
      if(button==GLUT_RIGHT_BUTTON || (button==GLUT_LEFT_BUTTON && mods&GLUT_ACTIVE_CTRL))
         mousemode=ZOOM;
      else if(button==GLUT_LEFT_BUTTON)
         mousemode=ROTATE;
   }
   oldmousex=x;
   oldmousey=y;
}

void handle_drag(int x, int y)
{
   switch(mousemode){
   case ROTATE:
      heading+=0.007*(oldmousex-x);
      if(heading<-M_PI) heading+=2*M_PI;
      else if(heading>M_PI) heading-=2*M_PI;
      pitch+=0.007*(oldmousey-y);
      if(pitch<-.48*M_PI) pitch=-.48*M_PI;
      else if(pitch>.48*M_PI) pitch=.48*M_PI;
      break;
   case ZOOM:
      camdist*=pow(1.01, y-oldmousey);
      nearclip=NEARCLIP_FACTOR*camdist;
      farclip=FARCLIP_FACTOR*camdist;
      break;
   case MOVETARGET:
      target[0]-=(0.002*camdist)*cos(heading)*(oldmousex-x);
      target[1]-=(0.002*camdist)*(oldmousey-y);
      target[2]-=(0.002*camdist)*sin(heading)*(oldmousex-x);
      break;
   case DONOTHING:
      break;
   }
   oldmousex=x;
   oldmousey=y;
   compute_camera_position();
   glutPostRedisplay();
}

void handle_idle()
{
   float new_time=clock()/(float)CLOCKS_PER_SEC;
   float delta_time=new_time-old_time;
   int target_turned_off=0;
   if(delta_time<0) delta_time=0; // this can happen due to roll-over
   old_time=new_time;
   if(view_target_timer>0){
      view_target_timer-=delta_time;
      if(view_target_timer<=0){
         target_turned_off=1;
         view_target_timer=0;
      }
   }
   if(target_turned_off)
      glutPostRedisplay();
}

void toggle_fullscreen()
{
   if(fullscreen_mode){
      fullscreen_mode=0;
      width=saved_width;
      height=saved_height;
      glutReshapeWindow(saved_width, saved_height);
   }else{
      fullscreen_mode=1;
      saved_width=width;
      saved_height=height;
      glutFullScreen();
   }
}

void compute_camera_position()
{
   float ch=cos(heading), sh=sin(heading), cp=cos(pitch), sp=sin(pitch);
   campos[0]=target[0]-camdist*sh*cp;
   campos[1]=target[1]-camdist*sp;
   campos[2]=target[2]-camdist*ch*cp;
}

void display_ground()
{
   glBegin(GL_QUADS);
   glNormal3f(0, 1, 0);
   glVertex3f(-3, 0, 3);
   glVertex3f(3, 0, 3);
   glVertex3f(3, 0, -3);
   glVertex3f(-3, 0, -3);
   glEnd();
}

void display_mesh(const TriMesh *mesh)
{
   int i;
   glBegin(GL_TRIANGLES);
   for(i=0; i<mesh->nt; ++i){
      glNormal3fv(mesh->vnormals+3*mesh->tri[3*i]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i]);
      glNormal3fv(mesh->vnormals+3*mesh->tri[3*i+1]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i+1]);
      glNormal3fv(mesh->vnormals+3*mesh->tri[3*i+2]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i+2]);
   }
   glEnd();
}

void display_target()
{
   float logcd=log(camdist)/log(20.);
   float smallsize=pow(20.0, floor(logcd-.5));
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(target[0], target[1], target[2]);
   glColor3f(1, 1, 1);
   glutWireCube(smallsize);
   glutWireCube(20*smallsize);
   glPopMatrix();
}

void save_screen(const char *filename)
{
   FILE *fp;
   GLubyte *image_buffer;

   // get the pixels
   image_buffer=malloc(3*width*height*sizeof(GLubyte));
   glReadBuffer(GL_FRONT);
   glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image_buffer);
  
   // write them to the file
   fp=fopen(filename, "wb");
   if(fp){
      int i;
      fprintf(fp, "P6\n%d %d 255\n", width, height);
      for(i=1; i<=height; ++i){
         fwrite(image_buffer+3*width*(height-i), 1, 3*width, fp);
      }
   }
   
   // clean up
   free(image_buffer);
}

